from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    ClassVar,
    TypeGuard,
)

from cross_web import (
    AsyncFlaskHTTPRequestAdapter,
    FlaskHTTPRequestAdapter,
    HTTPException,
)

from flask import Request, Response, render_template_string, request
from flask.views import View
from strawberry.http.async_base_view import AsyncBaseHTTPView
from strawberry.http.sync_base_view import SyncBaseHTTPView
from strawberry.http.typevars import Context, RootValue

if TYPE_CHECKING:
    from flask.typing import ResponseReturnValue

    from strawberry.http import GraphQLHTTPResponse
    from strawberry.http.ides import GraphQL_IDE
    from strawberry.schema.base import BaseSchema


class BaseGraphQLView:
    graphql_ide: GraphQL_IDE | None

    def __init__(
        self,
        schema: BaseSchema,
        graphql_ide: GraphQL_IDE | None = "graphiql",
        allow_queries_via_get: bool = True,
        multipart_uploads_enabled: bool = False,
    ) -> None:
        self.schema = schema
        self.allow_queries_via_get = allow_queries_via_get
        self.multipart_uploads_enabled = multipart_uploads_enabled
        self.graphql_ide = graphql_ide

    def create_response(
        self,
        response_data: GraphQLHTTPResponse | list[GraphQLHTTPResponse],
        sub_response: Response,
    ) -> Response:
        sub_response.set_data(self.encode_json(response_data))  # type: ignore

        return sub_response


class GraphQLView(
    BaseGraphQLView,
    SyncBaseHTTPView[Request, Response, Response, Context, RootValue],
    View,
):
    methods: ClassVar[list[str]] = ["GET", "POST"]
    allow_queries_via_get: bool = True
    request_adapter_class = FlaskHTTPRequestAdapter

    def get_context(self, request: Request, response: Response) -> Context:
        return {"request": request, "response": response}  # type: ignore

    def get_root_value(self, request: Request) -> RootValue | None:
        return None

    def get_sub_response(self, request: Request) -> Response:
        return Response(status=200, content_type="application/json")

    def dispatch_request(self) -> ResponseReturnValue:
        try:
            return self.run(request=request)
        except HTTPException as e:
            return Response(
                response=e.reason,
                status=e.status_code,
                content_type="text/plain",
            )

    def render_graphql_ide(self, request: Request) -> Response:
        return render_template_string(self.graphql_ide_html)  # type: ignore


class AsyncGraphQLView(
    BaseGraphQLView,
    AsyncBaseHTTPView[
        Request, Response, Response, Request, Response, Context, RootValue
    ],
    View,
):
    methods: ClassVar[list[str]] = ["GET", "POST"]
    allow_queries_via_get: bool = True
    request_adapter_class = AsyncFlaskHTTPRequestAdapter

    async def get_context(self, request: Request, response: Response) -> Context:
        return {"request": request, "response": response}  # type: ignore

    async def get_root_value(self, request: Request) -> RootValue | None:
        return None

    async def get_sub_response(self, request: Request) -> Response:
        return Response(status=200, content_type="application/json")

    async def dispatch_request(self) -> ResponseReturnValue:  # type: ignore
        try:
            return await self.run(request=request)
        except HTTPException as e:
            return Response(
                response=e.reason,
                status=e.status_code,
                content_type="text/plain",
            )

    async def render_graphql_ide(self, request: Request) -> Response:
        content = render_template_string(self.graphql_ide_html)
        return Response(content, status=200, content_type="text/html")

    def is_websocket_request(self, request: Request) -> TypeGuard[Request]:
        return False

    async def pick_websocket_subprotocol(self, request: Request) -> str | None:
        raise NotImplementedError

    async def create_websocket_response(
        self, request: Request, subprotocol: str | None
    ) -> Response:
        raise NotImplementedError


__all__ = [
    "AsyncGraphQLView",
    "GraphQLView",
]



from .controller import (
    BaseContext,
    HTTPContextType,
    WebSocketContextType,
    make_graphql_controller,
)

__all__ = [
    "BaseContext",
    "HTTPContextType",
    "WebSocketContextType",
    "make_graphql_controller",
]



from __future__ import annotations

import asyncio
import dataclasses
from abc import ABC, abstractmethod
from asyncio import create_task, gather, get_event_loop
from asyncio.futures import Future
from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    TypeVar,
    overload,
)

from .exceptions import WrongNumberOfResultsReturned

if TYPE_CHECKING:
    from asyncio.events import AbstractEventLoop
    from collections.abc import (
        Awaitable,
        Callable,
        Hashable,
        Iterable,
        Mapping,
        Sequence,
    )


T = TypeVar("T")
K = TypeVar("K")


@dataclass
class LoaderTask(Generic[K, T]):
    key: K
    future: Future


@dataclass
class Batch(Generic[K, T]):
    tasks: list[LoaderTask] = dataclasses.field(default_factory=list)
    dispatched: bool = False
    _dispatch_task: asyncio.Task[None] | None = dataclasses.field(
        default=None, repr=False
    )

    def add_task(self, key: Any, future: Future) -> None:
        task = LoaderTask[K, T](key, future)
        self.tasks.append(task)
        future.add_done_callback(self._on_future_done)

    def _on_future_done(self, future: Future) -> None:
        """Cancel the dispatch task if all futures in the batch are cancelled."""
        if (
            future.cancelled()
            and self._dispatch_task is not None
            and not self._dispatch_task.done()
            and all(t.future.cancelled() for t in self.tasks)
        ):
            self._dispatch_task.cancel()

    def __len__(self) -> int:
        return len(self.tasks)


class AbstractCache(ABC, Generic[K, T]):
    @abstractmethod
    def get(self, key: K) -> Future[T] | None:
        pass

    @abstractmethod
    def set(self, key: K, value: Future[T]) -> None:
        pass

    @abstractmethod
    def delete(self, key: K) -> None:
        pass

    @abstractmethod
    def clear(self) -> None:
        pass


class DefaultCache(AbstractCache[K, T]):
    def __init__(self, cache_key_fn: Callable[[K], Hashable] | None = None) -> None:
        self.cache_key_fn: Callable[[K], Hashable] = (
            cache_key_fn if cache_key_fn is not None else lambda x: x
        )
        self.cache_map: dict[Hashable, Future[T]] = {}

    def get(self, key: K) -> Future[T] | None:
        return self.cache_map.get(self.cache_key_fn(key))

    def set(self, key: K, value: Future[T]) -> None:
        self.cache_map[self.cache_key_fn(key)] = value

    def delete(self, key: K) -> None:
        del self.cache_map[self.cache_key_fn(key)]

    def clear(self) -> None:
        self.cache_map.clear()


class DataLoader(Generic[K, T]):
    batch: Batch[K, T] | None = None
    cache: bool = False
    cache_map: AbstractCache[K, T]

    @overload
    def __init__(
        self,
        # any BaseException is rethrown in 'load', so should be excluded from the T type
        load_fn: Callable[[list[K]], Awaitable[Sequence[T | BaseException]]],
        max_batch_size: int | None = None,
        cache: bool = True,
        loop: AbstractEventLoop | None = None,
        cache_map: AbstractCache[K, T] | None = None,
        cache_key_fn: Callable[[K], Hashable] | None = None,
    ) -> None: ...

    # fallback if load_fn is untyped and there's no other info for inference
    @overload
    def __init__(
        self: DataLoader[K, Any],
        load_fn: Callable[[list[K]], Awaitable[list[Any]]],
        max_batch_size: int | None = None,
        cache: bool = True,
        loop: AbstractEventLoop | None = None,
        cache_map: AbstractCache[K, T] | None = None,
        cache_key_fn: Callable[[K], Hashable] | None = None,
    ) -> None: ...

    def __init__(
        self,
        load_fn: Callable[[list[K]], Awaitable[Sequence[T | BaseException]]],
        max_batch_size: int | None = None,
        cache: bool = True,
        loop: AbstractEventLoop | None = None,
        cache_map: AbstractCache[K, T] | None = None,
        cache_key_fn: Callable[[K], Hashable] | None = None,
    ):
        self.load_fn = load_fn
        self.max_batch_size = max_batch_size

        self._loop = loop

        self.cache = cache

        if self.cache:
            self.cache_map = (
                DefaultCache(cache_key_fn) if cache_map is None else cache_map
            )

    @property
    def loop(self) -> AbstractEventLoop:
        if self._loop is None:
            self._loop = get_event_loop()

        return self._loop

    def load(self, key: K) -> Awaitable[T]:
        if self.cache:
            future = self.cache_map.get(key)

            if future and not future.cancelled():
                return future

        future = self.loop.create_future()

        if self.cache:
            self.cache_map.set(key, future)

        batch = get_current_batch(self)
        batch.add_task(key, future)

        return future

    def load_many(self, keys: Iterable[K]) -> Awaitable[list[T]]:
        return gather(*map(self.load, keys))

    def clear(self, key: K) -> None:
        if self.cache:
            self.cache_map.delete(key)

    def clear_many(self, keys: Iterable[K]) -> None:
        if self.cache:
            for key in keys:
                self.cache_map.delete(key)

    def clear_all(self) -> None:
        if self.cache:
            self.cache_map.clear()

    def prime(self, key: K, value: T, force: bool = False) -> None:
        self.prime_many({key: value}, force)

    def prime_many(self, data: Mapping[K, T], force: bool = False) -> None:
        # Populate the cache with the specified values
        if self.cache:
            for key, value in data.items():
                if not self.cache_map.get(key) or force:
                    future: Future = Future(loop=self.loop)
                    future.set_result(value)
                    self.cache_map.set(key, future)

        # For keys that are pending on the current batch, but the
        # batch hasn't started fetching yet: Remove it from the
        # batch and set to the specified value
        if self.batch is not None and not self.batch.dispatched:
            batch_updated = False
            for task in self.batch.tasks:
                if task.key in data:
                    batch_updated = True
                    task.future.set_result(data[task.key])
            if batch_updated:
                self.batch.tasks = [
                    task for task in self.batch.tasks if not task.future.done()
                ]


def should_create_new_batch(loader: DataLoader, batch: Batch) -> bool:
    return bool(
        batch.dispatched
        or (loader.max_batch_size and len(batch) >= loader.max_batch_size)
    )


def get_current_batch(loader: DataLoader) -> Batch:
    if loader.batch and not should_create_new_batch(loader, loader.batch):
        return loader.batch

    loader.batch = Batch()

    dispatch(loader, loader.batch)

    return loader.batch


def dispatch(loader: DataLoader, batch: Batch) -> None:
    def _schedule() -> None:
        batch._dispatch_task = create_task(dispatch_batch(loader, batch))
        batch._dispatch_task.add_done_callback(
            lambda _: setattr(batch, "_dispatch_task", None)
        )

    loader.loop.call_soon(_schedule)


async def dispatch_batch(loader: DataLoader, batch: Batch) -> None:
    batch.dispatched = True

    keys = [task.key for task in batch.tasks]
    if len(keys) == 0:
        # Ensure batch is not empty
        # Unlikely, but could happen if the tasks are
        # overridden with preset values
        return

    # Skip the batch entirely if all futures have already been cancelled.
    # This avoids calling load_fn (e.g. a database query) when no caller
    # is waiting for the results.
    if all(task.future.cancelled() for task in batch.tasks):
        return

    # TODO: check if load_fn return an awaitable and it is a list

    try:
        values = await loader.load_fn(keys)
        values = list(values)

        if len(values) != len(batch):
            raise WrongNumberOfResultsReturned(  # noqa: TRY301
                expected=len(batch), received=len(values)
            )

        for task, value in zip(batch.tasks, values, strict=True):
            # Trying to set_result in a cancelled future would raise
            # asyncio.exceptions.InvalidStateError
            if task.future.cancelled():
                continue

            if isinstance(value, BaseException):
                task.future.set_exception(value)
            else:
                task.future.set_result(value)
    except Exception as e:  # noqa: BLE001
        for task in batch.tasks:
            if task.future.cancelled():
                continue
            task.future.set_exception(e)


__all__ = [
    "AbstractCache",
    "Batch",
    "DataLoader",
    "DefaultCache",
    "LoaderTask",
    "dispatch",
    "dispatch_batch",
    "get_current_batch",
    "should_create_new_batch",
]



from __future__ import annotations

import re
import warnings
from decimal import Decimal
from typing import (
    TYPE_CHECKING,
    Any,
    cast,
)

from mypy.nodes import (
    ARG_OPT,
    ARG_POS,
    ARG_STAR2,
    MDEF,
    Argument,
    Block,
    CallExpr,
    CastExpr,
    FuncDef,
    IndexExpr,
    MemberExpr,
    NameExpr,
    PassStmt,
    SymbolTableNode,
    Var,
)
from mypy.plugin import (
    Plugin,
    SemanticAnalyzerPluginInterface,
)
from mypy.plugins.common import _get_argument, add_method
from mypy.semanal_shared import set_callable_name
from mypy.types import (
    AnyType,
    CallableType,
    Instance,
    NoneType,
    Type,
    TypeOfAny,
    TypeVarType,
    UnionType,
)
from mypy.typevars import fill_typevars
from mypy.util import get_unique_redefinition_name

if TYPE_CHECKING:
    from collections.abc import Callable

    from mypy.nodes import ClassDef, Expression
    from mypy.plugin import (
        CheckerPluginInterface,
        ClassDefContext,
    )

VERSION_RE = re.compile(r"(^0|^(?:[1-9][0-9]*))\.(0|(?:[1-9][0-9]*))")
FALLBACK_VERSION = Decimal("0.800")


class MypyVersion:
    VERSION: Decimal = FALLBACK_VERSION


class InvalidNodeTypeException(Exception):
    def __init__(self, node: Any) -> None:
        message = f"Invalid node type: {node!s}"
        self.message = message
        super().__init__(message)

    def __str__(self) -> str:
        return self.message


def _get_named_type(name: str, api: SemanticAnalyzerPluginInterface) -> Any:
    if "." in name:
        return api.named_type_or_none(name)
    return api.named_type(name)


def _get_type_for_expr(expr: Expression, api: SemanticAnalyzerPluginInterface) -> Type:
    if isinstance(expr, NameExpr):
        if expr.fullname:
            sym = api.lookup_fully_qualified_or_none(expr.fullname)
            if sym and isinstance(sym.node, Var):
                raise InvalidNodeTypeException(sym.node)
        return _get_named_type(expr.fullname or expr.name, api)

    if isinstance(expr, IndexExpr):
        type_ = _get_type_for_expr(expr.base, api)
        type_.args = (_get_type_for_expr(expr.index, api),)  # type: ignore
        return type_

    if isinstance(expr, MemberExpr):
        if expr.fullname:
            return _get_named_type(expr.fullname, api)
        raise InvalidNodeTypeException(expr)

    if isinstance(expr, CallExpr):
        if expr.analyzed:
            return _get_type_for_expr(expr.analyzed, api)
        raise InvalidNodeTypeException(expr)

    if isinstance(expr, CastExpr):
        return expr.type

    raise ValueError(f"Unsupported expression {type(expr)}")


def add_static_method_to_class(
    api: SemanticAnalyzerPluginInterface | CheckerPluginInterface,
    cls: ClassDef,
    name: str,
    args: list[Argument],
    return_type: Type,
    tvar_def: TypeVarType | None = None,
) -> None:
    info = cls.info

    if name in info.names:
        sym = info.names[name]
        if sym.plugin_generated and isinstance(sym.node, FuncDef):
            cls.defs.body.remove(sym.node)

    if Decimal("0.93") > MypyVersion.VERSION:
        function_type = api.named_type("__builtins__.function")  # type: ignore[union-attr]
    elif isinstance(api, SemanticAnalyzerPluginInterface):
        function_type = api.named_type("builtins.function")
    else:
        function_type = api.named_generic_type("builtins.function", [])

    arg_types, arg_names, arg_kinds = [], [], []
    for arg in args:
        assert arg.type_annotation, "All arguments must be fully typed."
        arg_types.append(arg.type_annotation)
        arg_names.append(arg.variable.name)
        arg_kinds.append(arg.kind)

    signature = CallableType(
        arg_types, arg_kinds, arg_names, return_type, function_type
    )
    if tvar_def:
        signature.variables = [tvar_def]  # type: ignore[assignment]

    func = FuncDef(name, args, Block([PassStmt()]))
    func.is_static = True
    func.info = info
    func.type = set_callable_name(signature, func)
    func._fullname = f"{info.fullname}.{name}"
    func.line = info.line

    if name in info.names:
        r_name = get_unique_redefinition_name(name, info.names)
        info.names[r_name] = info.names[name]

    info.names[name] = SymbolTableNode(MDEF, func, plugin_generated=True)
    info.defn.defs.body.append(func)


def strawberry_pydantic_class_callback(ctx: ClassDefContext) -> None:
    model_expression = _get_argument(call=cast("CallExpr", ctx.reason), name="model")
    if model_expression is None:
        ctx.api.fail("model argument in decorator failed to be parsed", ctx.reason)
        return

    # Add __init__ with **kwargs
    init_args = [Argument(Var("kwargs"), AnyType(TypeOfAny.explicit), None, ARG_STAR2)]
    add_method(ctx, "__init__", init_args, NoneType())

    try:
        model_type = cast("Instance", _get_type_for_expr(model_expression, ctx.api))
    except (InvalidNodeTypeException, ValueError):
        model_type = None

    if model_type is not None:
        # Add to_pydantic() -> ModelType
        if "to_pydantic" not in ctx.cls.info.names:
            to_pydantic_args = [
                Argument(Var("kwargs"), AnyType(TypeOfAny.explicit), None, ARG_STAR2)
            ]
            add_method(ctx, "to_pydantic", to_pydantic_args, model_type)

        # Add from_pydantic(instance, extra) as static method
        if "from_pydantic" not in ctx.cls.info.names:
            model_argument = Argument(
                variable=Var(name="instance", type=model_type),
                type_annotation=model_type,
                initializer=None,
                kind=ARG_POS,
            )
            extra_type = ctx.api.named_type(
                "builtins.dict",
                [ctx.api.named_type("builtins.str"), AnyType(TypeOfAny.explicit)],
            )
            extra_argument = Argument(
                variable=Var(name="extra", type=UnionType([NoneType(), extra_type])),
                type_annotation=UnionType([NoneType(), extra_type]),
                initializer=None,
                kind=ARG_OPT,
            )

            add_static_method_to_class(
                ctx.api,
                ctx.cls,
                name="from_pydantic",
                args=[model_argument, extra_argument],
                return_type=fill_typevars(ctx.cls.info),
            )
    # Fallback: add methods with Any types
    else:
        if "to_pydantic" not in ctx.cls.info.names:
            to_pydantic_args = [
                Argument(Var("kwargs"), AnyType(TypeOfAny.explicit), None, ARG_STAR2)
            ]
            add_method(
                ctx, "to_pydantic", to_pydantic_args, AnyType(TypeOfAny.explicit)
            )

        if "from_pydantic" not in ctx.cls.info.names:
            fallback_model_argument = Argument(
                variable=Var(name="instance", type=AnyType(TypeOfAny.explicit)),
                type_annotation=AnyType(TypeOfAny.explicit),
                initializer=None,
                kind=ARG_POS,
            )
            fallback_extra_argument = Argument(
                variable=Var(name="extra", type=AnyType(TypeOfAny.explicit)),
                type_annotation=AnyType(TypeOfAny.explicit),
                initializer=None,
                kind=ARG_OPT,
            )
            add_static_method_to_class(
                ctx.api,
                ctx.cls,
                name="from_pydantic",
                args=[fallback_model_argument, fallback_extra_argument],
                return_type=fill_typevars(ctx.cls.info),
            )


class StrawberryPlugin(Plugin):
    def get_class_decorator_hook(
        self, fullname: str
    ) -> Callable[[ClassDefContext], None] | None:
        if self._is_strawberry_pydantic_decorator(fullname):
            return strawberry_pydantic_class_callback
        return None

    _PYDANTIC_DECORATORS = frozenset(
        {
            "strawberry.experimental.pydantic.object_type.type",
            "strawberry.experimental.pydantic.object_type.input",
            "strawberry.experimental.pydantic.object_type.interface",
            "strawberry.experimental.pydantic.error_type.error_type",
            "strawberry.experimental.pydantic.type",
            "strawberry.experimental.pydantic.input",
            "strawberry.experimental.pydantic.error_type",
        }
    )

    def _is_strawberry_pydantic_decorator(self, fullname: str) -> bool:
        return fullname in self._PYDANTIC_DECORATORS


def plugin(version: str) -> type[StrawberryPlugin]:
    match = VERSION_RE.match(version)
    if match:
        MypyVersion.VERSION = Decimal(".".join(match.groups()))
    else:
        MypyVersion.VERSION = FALLBACK_VERSION
        warnings.warn(
            f"Mypy version {version} could not be parsed. Reverting to v0.800",
            stacklevel=1,
        )
    return StrawberryPlugin



from __future__ import annotations

import libcst as cst
import libcst.matchers as m
from libcst.codemod import CodemodContext, VisitorBasedCodemodCommand
from libcst.codemod.visitors import AddImportsVisitor

# Mapping from old scalar wrapper names to their replacements
# Format: old_name -> (module, import_name, usage_name)
# If usage_name is None, it means we use module.import_name
SCALAR_REPLACEMENTS: dict[str, tuple[str, str, str | None]] = {
    "Date": ("datetime", "date", "datetime.date"),
    "DateTime": ("datetime", "datetime", "datetime.datetime"),
    "Time": ("datetime", "time", "datetime.time"),
    "Decimal": ("decimal", "Decimal", None),
    "UUID": ("uuid", "UUID", None),
}


class ReplaceScalarWrappers(VisitorBasedCodemodCommand):
    DESCRIPTION: str = (
        "Replaces deprecated scalar wrapper imports from "
        "strawberry.schema.types.base_scalars with their actual Python types. "
        "For example: Date -> datetime.date, UUID -> uuid.UUID"
    )
    METADATA_DEPENDENCIES = (cst.metadata.ParentNodeProvider,)

    def __init__(self, context: CodemodContext) -> None:
        super().__init__(context)
        # Track which scalars are imported and need replacement
        self._imported_scalars: dict[str, str] = {}  # alias -> original_name

    def visit_ImportFrom(self, node: cst.ImportFrom) -> bool | None:  # noqa: N802
        """Track imports from strawberry.schema.types.base_scalars."""
        if not m.matches(
            node,
            m.ImportFrom(
                module=m.Attribute(
                    value=m.Attribute(
                        value=m.Attribute(
                            value=m.Name("strawberry"),
                            attr=m.Name("schema"),
                        ),
                        attr=m.Name("types"),
                    ),
                    attr=m.Name("base_scalars"),
                )
            ),
        ):
            return True

        if isinstance(node.names, cst.ImportStar):
            return True

        for name in node.names:
            if isinstance(name, cst.ImportAlias):
                original_name = (
                    name.name.value if isinstance(name.name, cst.Name) else None
                )
                if original_name and original_name in SCALAR_REPLACEMENTS:
                    # Get the alias if present, otherwise use the original name
                    alias = (
                        name.asname.name.value
                        if name.asname and isinstance(name.asname.name, cst.Name)
                        else original_name
                    )
                    self._imported_scalars[alias] = original_name

        return True

    def leave_ImportFrom(  # noqa: N802
        self, original_node: cst.ImportFrom, updated_node: cst.ImportFrom
    ) -> cst.ImportFrom | cst.RemovalSentinel:
        """Remove or update imports from strawberry.schema.types.base_scalars."""
        if not m.matches(
            original_node,
            m.ImportFrom(
                module=m.Attribute(
                    value=m.Attribute(
                        value=m.Attribute(
                            value=m.Name("strawberry"),
                            attr=m.Name("schema"),
                        ),
                        attr=m.Name("types"),
                    ),
                    attr=m.Name("base_scalars"),
                )
            ),
        ):
            return updated_node

        if isinstance(original_node.names, cst.ImportStar):
            return updated_node

        # Filter out the scalar wrappers we're replacing
        remaining_names: list[cst.ImportAlias] = [
            name
            for name in original_node.names
            if (
                isinstance(name, cst.ImportAlias)
                and isinstance(name.name, cst.Name)
                and name.name.value not in SCALAR_REPLACEMENTS
            )
        ]

        # Add the replacement imports for all imported scalars
        for original_name in self._imported_scalars.values():
            module, import_name, usage_name = SCALAR_REPLACEMENTS[original_name]

            if usage_name and usage_name.startswith("datetime."):
                # For datetime types, we import the module
                AddImportsVisitor.add_needed_import(self.context, module)
            else:
                # For other types, we import the specific name
                AddImportsVisitor.add_needed_import(self.context, module, import_name)

        if not remaining_names:
            # Remove the entire import statement
            return cst.RemovalSentinel.REMOVE

        # Update the import to only include non-scalar names
        return updated_node.with_changes(
            names=remaining_names,
        )

    def leave_Name(  # noqa: N802
        self, original_node: cst.Name, updated_node: cst.Name
    ) -> cst.BaseExpression:
        """Replace usages of imported scalar wrappers with their actual types."""
        if original_node.value not in self._imported_scalars:
            return updated_node

        # Don't replace names that are part of import statements
        try:
            parent = self.get_metadata(cst.metadata.ParentNodeProvider, original_node)
            if isinstance(parent, (cst.ImportAlias, cst.AsName)):
                return updated_node
        except KeyError:
            pass

        original_name = self._imported_scalars[original_node.value]
        _module, import_name, usage_name = SCALAR_REPLACEMENTS[original_name]

        if usage_name:
            # Return module.name (e.g., datetime.date)
            parts = usage_name.split(".")
            if len(parts) == 2:
                return cst.Attribute(
                    value=cst.Name(parts[0]),
                    attr=cst.Name(parts[1]),
                )

        # Return just the name (e.g., UUID, Decimal)
        return cst.Name(import_name)



from .handlers.base import ChannelsConsumer
from .handlers.http_handler import (
    ChannelsRequest,
    GraphQLHTTPConsumer,
    SyncGraphQLHTTPConsumer,
)
from .handlers.ws_handler import GraphQLWSConsumer
from .router import GraphQLProtocolTypeRouter

__all__ = [
    "ChannelsConsumer",
    "ChannelsRequest",
    "GraphQLHTTPConsumer",
    "GraphQLProtocolTypeRouter",
    "GraphQLWSConsumer",
    "SyncGraphQLHTTPConsumer",
]



from __future__ import annotations as _

import asyncio
import inspect
import sys
from functools import cached_property
from inspect import isasyncgenfunction
from typing import (
    TYPE_CHECKING,
    Annotated,
    Any,
    ForwardRef,
    Generic,
    NamedTuple,
    TypeVar,
    cast,
    get_origin,
)
from typing_extensions import Protocol

from strawberry.annotation import StrawberryAnnotation
from strawberry.exceptions import (
    ConflictingArgumentsError,
    MissingArgumentsAnnotationsError,
)
from strawberry.parent import StrawberryParent, resolve_parent_forward_arg
from strawberry.types.arguments import StrawberryArgument
from strawberry.types.base import StrawberryType, has_object_definition
from strawberry.types.info import Info
from strawberry.utils.typing import type_has_annotation

if TYPE_CHECKING:
    import builtins
    from collections.abc import Callable, Mapping


class Parameter(inspect.Parameter):
    def __hash__(self) -> int:
        """Override to exclude default value from hash.

        This adds compatibility for using unhashable default values in resolvers such as
        list and dict. The present use-case is limited to analyzing parameters from one
        resolver. Therefore, the name, kind, and annotation combination are guaranteed
        to be unique since two arguments cannot have the same name in a callable.

        Furthermore, even though it is not currently a use-case to collect parameters
        from different resolvers, the likelihood of collision from having the same hash
        value but different defaults is mitigated by Python invoking the
        :py:meth:`__eq__` method if two items have the same hash. See the verification
        of this behavior in the `test_parameter_hash_collision` test.
        """
        return hash((self.name, self.kind, self.annotation))


class Signature(inspect.Signature):
    _parameter_cls = Parameter


class ReservedParameterSpecification(Protocol):
    def find(
        self,
        parameters: tuple[inspect.Parameter, ...],
        resolver: StrawberryResolver[Any],
    ) -> inspect.Parameter | None:
        """Finds the reserved parameter from ``parameters``."""

    def __hash__(self) -> int: ...


class ReservedName(NamedTuple):
    name: str

    def find(
        self,
        parameters: tuple[inspect.Parameter, ...],
        resolver: StrawberryResolver[Any],
    ) -> inspect.Parameter | None:
        del resolver
        return next((p for p in parameters if p.name == self.name), None)


class ReservedNameBoundParameter(NamedTuple):
    name: str

    def find(
        self,
        parameters: tuple[inspect.Parameter, ...],
        resolver: StrawberryResolver[Any],
    ) -> inspect.Parameter | None:
        del resolver
        if parameters:  # Add compatibility for resolvers with no arguments
            first_parameter = parameters[0]
            return first_parameter if first_parameter.name == self.name else None
        return None


class ReservedType(NamedTuple):
    """Define a reserved type by name or by type.

    To preserve backwards-comaptibility, if an annotation was defined but does not match
    :attr:`type`, then the name is used as a fallback if available.

    :attr:`alias` is the public name exposed in the `strawberry` module, if it differs
    from the internal class name (e.g. "Parent" for StrawberryParent).
    """

    name: str | None
    type: type
    alias: str | None = None

    def find(
        self,
        parameters: tuple[inspect.Parameter, ...],
        resolver: StrawberryResolver[Any],
    ) -> inspect.Parameter | None:
        # Go through all the types even after we've found one so we can
        # give a helpful error message if someone uses the type more than once.
        type_parameters = []
        for parameter in parameters:
            annotation = resolver.strawberry_annotations[parameter]
            if isinstance(annotation, StrawberryAnnotation):
                try:
                    evaled_annotation: Any = annotation.evaluate()
                except NameError:
                    # If we fail to evaluate, check if the raw annotation string
                    # matches this reserved type. This handles cases where types
                    # are imported under TYPE_CHECKING or use forward references
                    # like "strawberry.Info".
                    raw = annotation.annotation
                    if isinstance(raw, str):
                        evaled_annotation = raw
                    else:
                        # Try resolve_parent_forward_arg for Parent types
                        evaled_annotation = resolve_parent_forward_arg(raw)

                if self.is_reserved_type(evaled_annotation):
                    type_parameters.append(parameter)

        if len(type_parameters) > 1:
            raise ConflictingArgumentsError(
                resolver, [parameter.name for parameter in type_parameters]
            )

        if type_parameters:
            return type_parameters[0]

        return None

    def is_reserved_type(self, other: builtins.type | str | ForwardRef) -> bool:
        # Handle string forward references (e.g., "strawberry.Info", "Info[Context, None]")
        # This occurs when annotations can't be resolved at runtime, such as when
        # the type is imported under TYPE_CHECKING or uses `from __future__ import annotations`
        if isinstance(other, str):
            return self._is_reserved_type_str(other)

        if isinstance(other, ForwardRef):
            return self._is_reserved_type_str(other.__forward_arg__)

        # Handle TypeAliasType (Python 3.12+ `type X = ...` syntax)
        # We need to unwrap the alias to get the actual type
        if hasattr(other, "__value__"):
            other = other.__value__

        origin = cast("type", get_origin(other)) or other
        if origin is Annotated:
            # Handle annotated arguments such as Private[str] and DirectiveValue[str]
            return type_has_annotation(other, self.type)

        # Handle both concrete and generic types (i.e. Info, and Info[X, Y])
        return (
            issubclass(origin, self.type)
            if isinstance(origin, type)
            else origin is self.type
        )

    def _is_reserved_type_str(self, annotation: str) -> bool:
        type_name = self.type.__name__
        module = self.type.__module__

        base_annotation = annotation.split("[", maxsplit=1)[0].strip()
        # Only match qualified names (strawberry.X or full module path)
        valid_names = {
            f"strawberry.{type_name}",  # e.g. "strawberry.Info"
            f"{module}.{type_name}",  # e.g. "strawberry.types.info.Info"
        }

        # Add alias if provided (e.g. "Parent" for StrawberryParent)
        if self.alias:
            valid_names.add(f"strawberry.{self.alias}")

        return base_annotation in valid_names


SELF_PARAMSPEC = ReservedNameBoundParameter("self")
CLS_PARAMSPEC = ReservedNameBoundParameter("cls")
ROOT_PARAMSPEC = ReservedName("root")
INFO_PARAMSPEC = ReservedType("info", Info)
PARENT_PARAMSPEC = ReservedType(name=None, type=StrawberryParent, alias="Parent")

T = TypeVar("T")

# in python >= 3.12 coroutine functions are market using inspect.markcoroutinefunction,
# which should be checked with inspect.iscoroutinefunction instead of asyncio.iscoroutinefunction
if hasattr(inspect, "markcoroutinefunction"):
    iscoroutinefunction = inspect.iscoroutinefunction
else:
    iscoroutinefunction = asyncio.iscoroutinefunction  # type: ignore[assignment]


class StrawberryResolver(Generic[T]):
    RESERVED_PARAMSPEC: tuple[ReservedParameterSpecification, ...] = (
        SELF_PARAMSPEC,
        CLS_PARAMSPEC,
        ROOT_PARAMSPEC,
        INFO_PARAMSPEC,
        PARENT_PARAMSPEC,
    )

    def __init__(
        self,
        func: Callable[..., T] | staticmethod | classmethod,
        *,
        description: str | None = None,
        type_override: StrawberryType | type | None = None,
    ) -> None:
        self.wrapped_func = func
        self._description = description
        self._type_override = type_override
        """Specify the type manually instead of calculating from wrapped func

        This is used when creating copies of types w/ generics
        """

    # TODO: Use this when doing the actual resolving? How to deal with async resolvers?
    def __call__(self, *args: str, **kwargs: Any) -> T:
        if not callable(self.wrapped_func):
            raise UncallableResolverError(self)
        return self.wrapped_func(*args, **kwargs)

    @cached_property
    def signature(self) -> inspect.Signature:
        return Signature.from_callable(self._unbound_wrapped_func, follow_wrapped=True)

    # TODO: find better name
    @cached_property
    def strawberry_annotations(
        self,
    ) -> dict[inspect.Parameter, StrawberryAnnotation | None]:
        return {
            p: (
                StrawberryAnnotation(p.annotation, namespace=self._namespace)
                if p.annotation is not inspect.Signature.empty
                else None
            )
            for p in self.signature.parameters.values()
        }

    @cached_property
    def reserved_parameters(
        self,
    ) -> dict[ReservedParameterSpecification, inspect.Parameter | None]:
        """Mapping of reserved parameter specification to parameter."""
        parameters = tuple(self.signature.parameters.values())
        return {spec: spec.find(parameters, self) for spec in self.RESERVED_PARAMSPEC}

    @cached_property
    def arguments(self) -> list[StrawberryArgument]:
        """Resolver arguments exposed in the GraphQL Schema."""
        root_parameter = self.reserved_parameters.get(ROOT_PARAMSPEC)
        parent_parameter = self.reserved_parameters.get(PARENT_PARAMSPEC)

        # TODO: Maybe use SELF_PARAMSPEC in the future? Right now
        # it would prevent some common pattern for integrations
        # (e.g. django) of typing the `root` parameters as the
        # type of the real object being used
        if (
            root_parameter is not None
            and parent_parameter is not None
            and root_parameter.name != parent_parameter.name
        ):
            raise ConflictingArgumentsError(
                self,
                [root_parameter.name, parent_parameter.name],
            )

        parameters = self.signature.parameters.values()
        reserved_parameters = set(self.reserved_parameters.values())
        missing_annotations: list[str] = []
        arguments: list[StrawberryArgument] = []
        user_parameters = (p for p in parameters if p not in reserved_parameters)

        for param in user_parameters:
            annotation = self.strawberry_annotations[param]
            if annotation is None:
                missing_annotations.append(param.name)
            else:
                argument = StrawberryArgument(
                    python_name=param.name,
                    graphql_name=None,
                    type_annotation=annotation,
                    default=param.default,
                )
                arguments.append(argument)
        if missing_annotations:
            raise MissingArgumentsAnnotationsError(self, missing_annotations)
        return arguments

    @cached_property
    def info_parameter(self) -> inspect.Parameter | None:
        return self.reserved_parameters.get(INFO_PARAMSPEC)

    @cached_property
    def root_parameter(self) -> inspect.Parameter | None:
        return self.reserved_parameters.get(ROOT_PARAMSPEC)

    @cached_property
    def self_parameter(self) -> inspect.Parameter | None:
        return self.reserved_parameters.get(SELF_PARAMSPEC)

    @cached_property
    def parent_parameter(self) -> inspect.Parameter | None:
        return self.reserved_parameters.get(PARENT_PARAMSPEC)

    @cached_property
    def name(self) -> str:
        # TODO: What to do if resolver is a lambda?
        return self._unbound_wrapped_func.__name__

    # TODO: consider deprecating
    @cached_property
    def annotations(self) -> dict[str, object]:
        """Annotations for the resolver.

        Does not include special args defined in `RESERVED_PARAMSPEC` (e.g. self, root,
        info)
        """
        reserved_parameters = self.reserved_parameters
        reserved_names = {p.name for p in reserved_parameters.values() if p is not None}

        annotations = self._unbound_wrapped_func.__annotations__
        return {
            name: annotation
            for name, annotation in annotations.items()
            if name not in reserved_names
        }

    @cached_property
    def type_annotation(self) -> StrawberryAnnotation | None:
        return_annotation = self.signature.return_annotation
        if return_annotation is inspect.Signature.empty:
            return None
        return StrawberryAnnotation(
            annotation=return_annotation, namespace=self._namespace
        )

    @property
    def type(self) -> StrawberryType | type | None:
        if self._type_override:
            return self._type_override
        if self.type_annotation is None:
            return None
        return self.type_annotation.resolve()

    @property
    def is_graphql_generic(self) -> bool:
        from strawberry.schema.compat import is_graphql_generic

        has_generic_arguments = any(
            argument.is_graphql_generic for argument in self.arguments
        )

        return has_generic_arguments or bool(
            self.type and is_graphql_generic(self.type)
        )

    @cached_property
    def is_async(self) -> bool:
        return iscoroutinefunction(self._unbound_wrapped_func) or isasyncgenfunction(
            self._unbound_wrapped_func
        )

    def copy_with(
        self, type_var_map: Mapping[str, StrawberryType | builtins.type]
    ) -> StrawberryResolver:
        type_override = None

        if self.type:
            if isinstance(self.type, StrawberryType):
                type_override = self.type.copy_with(type_var_map)
            elif has_object_definition(self.type):
                type_override = self.type.__strawberry_definition__.copy_with(
                    type_var_map,
                )

        other = type(self)(
            func=self.wrapped_func,
            description=self._description,
            type_override=type_override,
        )
        # Resolve generic arguments
        for argument in other.arguments:
            if (
                isinstance(argument.type, StrawberryType)
                and argument.type.is_graphql_generic
            ):
                argument.type_annotation = StrawberryAnnotation(
                    annotation=argument.type.copy_with(type_var_map),
                    namespace=argument.type_annotation.namespace,
                )
        return other

    @cached_property
    def _namespace(self) -> dict[str, Any]:
        return sys.modules[self._unbound_wrapped_func.__module__].__dict__

    @cached_property
    def _unbound_wrapped_func(self) -> Callable[..., T]:
        if isinstance(self.wrapped_func, (staticmethod, classmethod)):
            return self.wrapped_func.__func__

        return self.wrapped_func


class UncallableResolverError(Exception):
    def __init__(self, resolver: StrawberryResolver) -> None:
        message = (
            f"Attempted to call resolver {resolver} with uncallable function "
            f"{resolver.wrapped_func}"
        )
        super().__init__(message)


__all__ = ["StrawberryResolver"]



from __future__ import annotations

import asyncio
import logging
from contextlib import suppress
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    cast,
)

from graphql import GraphQLError, GraphQLSyntaxError

from strawberry.exceptions import ConnectionRejectionError
from strawberry.http.exceptions import (
    NonJsonMessageReceived,
    NonTextMessageReceived,
    WebSocketDisconnected,
)
from strawberry.http.typevars import Context, RootValue
from strawberry.schema.exceptions import CannotGetOperationTypeError
from strawberry.subscriptions.protocols.graphql_transport_ws.types import (
    CompleteMessage,
    ConnectionInitMessage,
    Message,
    NextMessagePayload,
    PingMessage,
    PongMessage,
    SubscribeMessage,
)
from strawberry.types.execution import ExecutionResult, PreExecutionError
from strawberry.types.unset import UnsetType
from strawberry.utils.aio import aclosing

if TYPE_CHECKING:
    from datetime import timedelta

    from strawberry.http.async_base_view import AsyncBaseHTTPView, AsyncWebSocketAdapter
    from strawberry.schema import BaseSchema
    from strawberry.schema.schema import StreamResult


class _OperationStreamClosed(Exception):
    """Raised when an operation closes the websocket instead of completing."""


class BaseGraphQLTransportWSHandler(Generic[Context, RootValue]):
    task_logger: logging.Logger = logging.getLogger("strawberry.ws.task")

    def __init__(
        self,
        view: AsyncBaseHTTPView[Any, Any, Any, Any, Any, Context, RootValue],
        websocket: AsyncWebSocketAdapter,
        context: Context,
        root_value: RootValue | None,
        schema: BaseSchema,
        connection_init_wait_timeout: timedelta,
        max_subscriptions_per_connection: int | None = None,
    ) -> None:
        self.view = view
        self.websocket = websocket
        self.context = context
        self.root_value = root_value
        self.schema = schema
        self.connection_init_wait_timeout = connection_init_wait_timeout
        self.max_subscriptions_per_connection = max_subscriptions_per_connection
        self.connection_init_timeout_task: asyncio.Task | None = None
        self.connection_init_received = False
        self.connection_acknowledged = False
        self.connection_timed_out = False
        self.operations: dict[str, Operation[Context, RootValue]] = {}

    async def handle(self) -> None:
        self.on_request_accepted()

        try:
            try:
                async for message in self.websocket.iter_json():
                    await self.handle_message(cast("Message", message))
            except NonTextMessageReceived:
                await self.handle_invalid_message("WebSocket message type must be text")
            except NonJsonMessageReceived:
                await self.handle_invalid_message(
                    "WebSocket message must be valid JSON"
                )
        except WebSocketDisconnected:
            pass
        finally:
            await self.shutdown()

    async def shutdown(self) -> None:
        if self.connection_init_timeout_task:
            self.connection_init_timeout_task.cancel()
            with suppress(asyncio.CancelledError):
                await self.connection_init_timeout_task

        tasks = [op.task for op in self.operations.values() if op.task]
        for operation_id in list(self.operations.keys()):
            await self.cleanup_operation(operation_id)
        # Let cancelled tasks finish their finally blocks before shutdown returns.
        # Per-task suppress survives parent cancellation; asyncio.gather would not.
        for task in tasks:
            with suppress(Exception, asyncio.CancelledError):
                await task

    def on_request_accepted(self) -> None:
        # handle_request should call this once it has sent the
        # websocket.accept() response to start the timeout.
        assert not self.connection_init_timeout_task
        self.connection_init_timeout_task = asyncio.create_task(
            self.handle_connection_init_timeout()
        )
        self.connection_init_timeout_task.add_done_callback(self._task_done)

    async def handle_connection_init_timeout(self) -> None:
        try:
            delay = self.connection_init_wait_timeout.total_seconds()
            await asyncio.sleep(delay=delay)

            if self.connection_init_received:
                return  # pragma: no cover

            self.connection_timed_out = True
            reason = "Connection initialisation timeout"
            await self.websocket.close(code=4408, reason=reason)
        except Exception as error:  # noqa: BLE001
            await self.handle_task_exception(error)  # pragma: no cover

    def _task_done(self, task: asyncio.Task) -> None:
        if task.cancelled():
            return
        # Retrieve exception to prevent "Task exception was never retrieved" warning.
        # Exceptions are already logged by handle_task_exception in the coroutines.
        task.exception()

    async def handle_task_exception(self, error: Exception) -> None:  # pragma: no cover
        self.task_logger.exception("Exception in worker task", exc_info=error)

    async def handle_message(self, message: Message) -> None:
        try:
            if message["type"] == "connection_init":
                await self.handle_connection_init(message)

            elif message["type"] == "ping":
                await self.handle_ping(message)

            elif message["type"] == "pong":
                await self.handle_pong(message)

            elif message["type"] == "subscribe":
                await self.handle_subscribe(message)

            elif message["type"] == "complete":
                await self.handle_complete(message)

            else:
                error_message = f"Unknown message type: {message['type']}"
                await self.handle_invalid_message(error_message)

        except KeyError:
            await self.handle_invalid_message("Failed to parse message")

    async def handle_connection_init(self, message: ConnectionInitMessage) -> None:
        if self.connection_timed_out:
            # No way to reliably exercise this case during testing
            return  # pragma: no cover

        if self.connection_init_timeout_task:
            self.connection_init_timeout_task.cancel()

        payload = message.get("payload", {})

        if not isinstance(payload, dict):
            await self.websocket.close(
                code=4400, reason="Invalid connection init payload"
            )
            return

        if self.connection_init_received:
            reason = "Too many initialisation requests"
            await self.websocket.close(code=4429, reason=reason)
            return

        self.connection_init_received = True

        if isinstance(self.context, dict):
            self.context["connection_params"] = payload
        elif hasattr(self.context, "connection_params"):
            self.context.connection_params = payload

        try:
            connection_ack_payload = await self.view.on_ws_connect(self.context)
        except ConnectionRejectionError:
            await self.websocket.close(code=4403, reason="Forbidden")
            return

        if isinstance(connection_ack_payload, UnsetType):
            await self.send_message({"type": "connection_ack"})
        else:
            await self.send_message(
                {"type": "connection_ack", "payload": connection_ack_payload}
            )

        self.connection_acknowledged = True

    async def handle_ping(self, message: PingMessage) -> None:
        await self.send_message({"type": "pong"})

    async def handle_pong(self, message: PongMessage) -> None:
        pass

    async def handle_subscribe(self, message: SubscribeMessage) -> None:
        if not self.connection_acknowledged:
            await self.websocket.close(code=4401, reason="Unauthorized")
            return

        if message["id"] in self.operations:
            reason = f"Subscriber for {message['id']} already exists"
            await self.websocket.close(code=4409, reason=reason)
            return

        # NOTE: this applies to all in-flight operations (queries and mutations
        # executed over WebSocket included), not only subscriptions.
        if (
            self.max_subscriptions_per_connection is not None
            and len(self.operations) >= self.max_subscriptions_per_connection
        ):
            error = GraphQLError("Subscription limit reached")
            await self.send_message(
                {
                    "id": message["id"],
                    "type": "error",
                    "payload": [error.formatted],
                }
            )
            return

        operation = Operation(
            self,
            message["id"],
            message["payload"]["query"],
            message["payload"].get("variables"),
            message["payload"].get("operationName"),
        )

        operation.task = asyncio.create_task(self.run_operation(operation))
        operation.task.add_done_callback(self._task_done)
        self.operations[message["id"]] = operation

    async def run_operation(self, operation: Operation[Context, RootValue]) -> None:
        """The operation task's top level method. Cleans-up and de-registers the operation once it is done."""
        try:
            result_source = await self.schema.stream(
                operation.query,
                variable_values=operation.variables,
                context_value=self.context,
                root_value=self.root_value,
                operation_name=operation.operation_name,
            )

            async with aclosing(result_source):
                await self._send_result_stream(operation, result_source)

            await operation.send_operation_message(
                CompleteMessage(id=operation.id, type="complete")
            )

        except _OperationStreamClosed:
            return
        except Exception as error:  # pragma: no cover
            await self.handle_task_exception(error)

            with suppress(Exception):
                await operation.send_operation_message(
                    {"id": operation.id, "type": "complete"}
                )

            self.operations.pop(operation.id, None)

            raise

    async def _send_result_stream(
        self,
        operation: Operation[Context, RootValue],
        result_source: StreamResult,
    ) -> None:
        is_first_result = True

        async for result in result_source:
            if is_first_result and isinstance(result, PreExecutionError):
                await self._handle_pre_execution_error(operation, result)
                return

            if not isinstance(result, ExecutionResult):
                await self._reject_incremental_delivery(operation)
                return

            await operation.send_next(result)
            is_first_result = False

    async def _handle_pre_execution_error(
        self,
        operation: Operation[Context, RootValue],
        result: PreExecutionError,
    ) -> None:
        assert result.errors
        if close_reason := self._get_pre_execution_close_reason(result.errors[0]):
            await self.websocket.close(code=4400, reason=close_reason)
            self.operations.pop(operation.id, None)
            raise _OperationStreamClosed

        await operation.send_initial_errors(result.errors)

    async def _reject_incremental_delivery(
        self, operation: Operation[Context, RootValue]
    ) -> None:
        # Incremental delivery (``@defer``/``@stream``) yields raw graphql-core
        # frames that the graphql-transport-ws ``next`` message has no wire
        # representation for. Reject the operation rather than send a malformed
        # or partial payload.
        await operation.send_initial_errors(
            [
                GraphQLError(
                    "Incremental delivery is not supported over graphql-transport-ws"
                )
            ]
        )

    def forget_id(self, id: str) -> None:
        # de-register the operation id making it immediately available
        # for reuse
        del self.operations[id]

    async def handle_complete(self, message: CompleteMessage) -> None:
        await self.cleanup_operation(operation_id=message["id"])

    async def handle_invalid_message(self, error_message: str) -> None:
        await self.websocket.close(code=4400, reason=error_message)

    @staticmethod
    def _get_pre_execution_close_reason(error: GraphQLError) -> str | None:
        if isinstance(error, GraphQLSyntaxError):
            return error.message

        if isinstance(error.original_error, CannotGetOperationTypeError):
            return error.original_error.as_http_error_reason()

        return None

    async def send_message(self, message: Message) -> None:
        await self.websocket.send_json(message)

    async def cleanup_operation(self, operation_id: str) -> None:
        if operation_id not in self.operations:
            return
        operation = self.operations.pop(operation_id)
        assert operation.task
        operation.task.cancel()
        # do not await the task here, lest we block the main
        # websocket handler Task.


class Operation(Generic[Context, RootValue]):
    """A class encapsulating a single operation with its id. Helps enforce protocol state transition."""

    __slots__ = [
        "completed",
        "handler",
        "id",
        "operation_name",
        "query",
        "task",
        "variables",
    ]

    def __init__(
        self,
        handler: BaseGraphQLTransportWSHandler[Context, RootValue],
        id: str,
        query: str,
        variables: dict[str, object] | None,
        operation_name: str | None,
    ) -> None:
        self.handler = handler
        self.id = id
        self.query = query
        self.variables = variables
        self.operation_name = operation_name
        self.completed = False
        self.task: asyncio.Task | None = None

    async def send_operation_message(self, message: Message) -> None:
        if self.completed:
            return
        if message["type"] == "complete" or message["type"] == "error":
            self.completed = True
            # de-register the operation _before_ sending the final message
            self.handler.forget_id(self.id)
        await self.handler.send_message(message)

    async def send_initial_errors(self, errors: list[GraphQLError]) -> None:
        # Initial errors see https://github.com/enisdenjo/graphql-ws/blob/master/PROTOCOL.md#error
        # "This can occur before execution starts,
        # usually due to validation errors, or during the execution of the request"
        await self.send_operation_message(
            {
                "id": self.id,
                "type": "error",
                "payload": [err.formatted for err in errors],
            }
        )

    async def send_next(self, execution_result: ExecutionResult) -> None:
        next_payload: NextMessagePayload = {"data": execution_result.data}

        if execution_result.errors:
            next_payload["errors"] = [err.formatted for err in execution_result.errors]

        if execution_result.extensions:
            next_payload["extensions"] = execution_result.extensions

        await self.send_operation_message(
            {"id": self.id, "type": "next", "payload": next_payload}
        )


__all__ = ["BaseGraphQLTransportWSHandler", "Operation"]



# Type alias for federation version as (major, minor) tuple
FederationVersion = tuple[int, int]

# Mapping from version string (e.g., "2.5") to version tuple (e.g., (2, 5))
FEDERATION_VERSIONS: dict[str, FederationVersion] = {
    "2.0": (2, 0),
    "2.1": (2, 1),
    "2.2": (2, 2),
    "2.3": (2, 3),
    "2.4": (2, 4),
    "2.5": (2, 5),
    "2.6": (2, 6),
    "2.7": (2, 7),
    "2.8": (2, 8),
    "2.9": (2, 9),
    "2.10": (2, 10),
    "2.11": (2, 11),
}


def parse_version(version_str: str) -> FederationVersion:
    """Parse a version string like '2.5' into a tuple (2, 5)."""
    return FEDERATION_VERSIONS[version_str]


def format_version(version: FederationVersion) -> str:
    """Format a version tuple (2, 5) into a string 'v2.5'."""
    major, minor = version
    return f"v{major}.{minor}"



import json
from collections.abc import Mapping
from functools import cached_property
from typing import Any, Generic
from typing_extensions import Protocol

from cross_web import HTTPException

from strawberry.http import GraphQLRequestData
from strawberry.http.ides import GraphQL_IDE, get_graphql_ide_html
from strawberry.http.types import HTTPMethod, QueryParams
from strawberry.schema.base import BaseSchema

from .streaming import HTTPStreamTransport, MultipartSubscriptionTransport
from .typevars import Request


class BaseRequestProtocol(Protocol):
    @property
    def query_params(self) -> Mapping[str, str | list[str] | None]: ...

    @property
    def method(self) -> HTTPMethod: ...

    @property
    def headers(self) -> Mapping[str, str]: ...


class BaseView(Generic[Request]):
    graphql_ide: GraphQL_IDE | None
    multipart_uploads_enabled: bool = False
    schema: BaseSchema
    stream_transport_classes: Mapping[str, type[HTTPStreamTransport]] = {
        MultipartSubscriptionTransport.protocol: MultipartSubscriptionTransport,
    }

    def should_render_graphql_ide(self, request: BaseRequestProtocol) -> bool:
        return (
            request.method == "GET"
            and request.query_params.get("query") is None
            and any(
                supported_header in request.headers.get("accept", "")
                for supported_header in ("text/html", "*/*")
            )
        )

    def is_request_allowed(self, request: BaseRequestProtocol) -> bool:
        return request.method in ("GET", "POST")

    def parse_json(self, data: str | bytes) -> Any:
        try:
            return self.decode_json(data)
        except json.JSONDecodeError as e:
            raise HTTPException(400, "Unable to parse request body as JSON") from e

    def decode_json(self, data: str | bytes) -> object:
        return json.loads(data)

    def encode_json(self, data: object) -> str | bytes:
        return json.dumps(data)

    def parse_query_params(self, params: QueryParams) -> dict[str, Any]:
        params = dict(params)

        if "variables" in params:
            variables = params["variables"]

            if variables:
                params["variables"] = self.parse_json(variables)

        if "extensions" in params:
            extensions = params["extensions"]

            if extensions:
                params["extensions"] = self.parse_json(extensions)

        return params

    @property
    def graphql_ide_html(self) -> str:
        return get_graphql_ide_html(graphql_ide=self.graphql_ide)

    @cached_property
    def _stream_transport_map(self) -> dict[str, HTTPStreamTransport]:
        return {
            protocol: transport_class()
            for protocol, transport_class in self.stream_transport_classes.items()
        }

    def _get_stream_transport(self, protocol: str) -> HTTPStreamTransport | None:
        return self._stream_transport_map.get(protocol)

    def _get_stream_transport_from_headers(
        self, headers: Mapping[str, str]
    ) -> HTTPStreamTransport | None:
        accept = next(
            (value for key, value in headers.items() if key.lower() == "accept"),
            "",
        )

        return next(
            (
                transport
                for transport in self._stream_transport_map.values()
                if transport.accepts(accept)
            ),
            None,
        )

    def _get_stream_transport_from_content_type(
        self, content_type: str, params: dict[str, str]
    ) -> HTTPStreamTransport | None:
        return next(
            (
                transport
                for transport in self._stream_transport_map.values()
                if transport.accepts_content_type(content_type, params)
            ),
            None,
        )

    def _validate_batch_request(
        self, request_data: list[GraphQLRequestData], protocol: str
    ) -> None:
        if self.schema.config.batching_config is None:
            raise HTTPException(400, "Batching is not enabled")

        if transport := self._get_stream_transport(protocol):
            raise HTTPException(400, transport.batching_error)

        if len(request_data) > self.schema.config.batching_config["max_operations"]:
            raise HTTPException(400, "Too many operations")


__all__ = ["BaseView"]



from __future__ import annotations

from typing import Any, TypeVar, overload

_T = TypeVar("_T", bound=object)

TYPE_CAST_ATTRIBUTE = "__as_strawberry_type__"


@overload
def cast(type_: type, obj: None) -> None: ...


@overload
def cast(type_: type, obj: _T) -> _T: ...


def cast(type_: type, obj: _T | None) -> _T | None:
    """Cast an object to given type.

    This is used to mark an object as a cast object, so that the type can be
    picked up when resolving unions/interfaces in case of ambiguity, which can
    happen when returning an alike object instead of an instance of the type
    (e.g. returning a Django, Pydantic or SQLAlchemy object)
    """
    if obj is None:
        return None

    setattr(obj, TYPE_CAST_ATTRIBUTE, type_)
    return obj


def get_strawberry_type_cast(obj: Any) -> type | None:
    """Get the type of a cast object."""
    return getattr(obj, TYPE_CAST_ATTRIBUTE, None)



from dataclasses import dataclass
from pathlib import Path


@dataclass
class ExceptionSource:
    path: Path
    code: str
    start_line: int
    end_line: int
    error_line: int
    error_column: int
    error_column_end: int

    @property
    def path_relative_to_cwd(self) -> Path:
        if self.path.is_absolute():
            return self.path.relative_to(Path.cwd())

        return self.path



from typing import Any

try:
    # import modules and objects from external strawberry-graphql-django
    # package so that it can be used through strawberry.django namespace
    from strawberry_django import *  # noqa: F403
except ModuleNotFoundError:
    import importlib

    def __getattr__(name: str) -> Any:
        # try to import submodule and raise exception only if it does not exist
        import_symbol = f"{__name__}.{name}"
        try:
            return importlib.import_module(import_symbol)
        except ModuleNotFoundError as e:
            raise AttributeError(
                f"Attempted import of {import_symbol} failed. Make sure to install the"
                "'strawberry-graphql-django' package to use the Strawberry Django "
                "extension API."
            ) from e



from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING

from .exception import StrawberryException
from .utils.source_finder import SourceFinder

if TYPE_CHECKING:
    from .exception_source import ExceptionSource


class MultipleStrawberryFieldsError(StrawberryException):
    def __init__(self, field_name: str, cls: type) -> None:
        self.cls = cls
        self.field_name = field_name

        self.message = (
            f"Annotation for field `{field_name}` on type `{cls.__name__}` "
            "cannot have multiple `strawberry.field`s"
        )
        self.rich_message = (
            f"Field `[underline]{self.field_name}[/]` on type "
            f"`[underline]{self.cls.__name__}[/]` cannot have multiple "
            "`strawberry.field`s"
        )
        self.annotation_message = "field with multiple strawberry.field annotations"
        self.suggestion = (
            "To fix this error you should use only one `strawberry.field()` "
            "in the `Annotated` type annotation."
        )

        super().__init__(self.message)

    @cached_property
    def exception_source(self) -> ExceptionSource | None:
        if self.cls is None:
            return None  # pragma: no cover

        source_finder = SourceFinder()

        return source_finder.find_class_attribute_from_object(self.cls, self.field_name)



from __future__ import annotations

import dataclasses
from typing import (
    TYPE_CHECKING,
    Any,
    Literal,
    overload,
)

from strawberry.types.field import (
    _RESOLVER_TYPE,
    _RESOLVER_TYPE_ASYNC,
    _RESOLVER_TYPE_SYNC,
    StrawberryField,
    T,
    field,
)

if TYPE_CHECKING:
    from collections.abc import Callable, Mapping, Sequence
    from typing import Literal

    from strawberry.extensions.field_extension import FieldExtension
    from strawberry.permission import BasePermission

# NOTE: we are separating the sync and async resolvers because using both
# in the same function will cause mypy to raise an error. Not sure if it is a bug


@overload
def mutation(
    *,
    resolver: _RESOLVER_TYPE_ASYNC[T],
    name: str | None = None,
    description: str | None = None,
    init: Literal[False] = False,
    permission_classes: list[type[BasePermission]] | None = None,
    deprecation_reason: str | None = None,
    default: Any = dataclasses.MISSING,
    default_factory: Callable[..., object] | object = dataclasses.MISSING,
    metadata: Mapping[Any, Any] | None = None,
    directives: Sequence[object] | None = (),
    extensions: list[FieldExtension] | None = None,
    graphql_type: Any | None = None,
) -> T: ...


@overload
def mutation(
    *,
    resolver: _RESOLVER_TYPE_SYNC[T],
    name: str | None = None,
    description: str | None = None,
    init: Literal[False] = False,
    permission_classes: list[type[BasePermission]] | None = None,
    deprecation_reason: str | None = None,
    default: Any = dataclasses.MISSING,
    default_factory: Callable[..., object] | object = dataclasses.MISSING,
    metadata: Mapping[Any, Any] | None = None,
    directives: Sequence[object] | None = (),
    extensions: list[FieldExtension] | None = None,
    graphql_type: Any | None = None,
) -> T: ...


@overload
def mutation(
    *,
    name: str | None = None,
    description: str | None = None,
    init: Literal[True] = True,
    permission_classes: list[type[BasePermission]] | None = None,
    deprecation_reason: str | None = None,
    default: Any = dataclasses.MISSING,
    default_factory: Callable[..., object] | object = dataclasses.MISSING,
    metadata: Mapping[Any, Any] | None = None,
    directives: Sequence[object] | None = (),
    extensions: list[FieldExtension] | None = None,
    graphql_type: Any | None = None,
) -> Any: ...


@overload
def mutation(
    resolver: _RESOLVER_TYPE_ASYNC[T],
    *,
    name: str | None = None,
    description: str | None = None,
    permission_classes: list[type[BasePermission]] | None = None,
    deprecation_reason: str | None = None,
    default: Any = dataclasses.MISSING,
    default_factory: Callable[..., object] | object = dataclasses.MISSING,
    metadata: Mapping[Any, Any] | None = None,
    directives: Sequence[object] | None = (),
    extensions: list[FieldExtension] | None = None,
    graphql_type: Any | None = None,
) -> StrawberryField: ...


@overload
def mutation(
    resolver: _RESOLVER_TYPE_SYNC[T],
    *,
    name: str | None = None,
    description: str | None = None,
    permission_classes: list[type[BasePermission]] | None = None,
    deprecation_reason: str | None = None,
    default: Any = dataclasses.MISSING,
    default_factory: Callable[..., object] | object = dataclasses.MISSING,
    metadata: Mapping[Any, Any] | None = None,
    directives: Sequence[object] | None = (),
    extensions: list[FieldExtension] | None = None,
    graphql_type: Any | None = None,
) -> StrawberryField: ...


def mutation(
    resolver: _RESOLVER_TYPE[Any] | None = None,
    *,
    name: str | None = None,
    description: str | None = None,
    permission_classes: list[type[BasePermission]] | None = None,
    deprecation_reason: str | None = None,
    default: Any = dataclasses.MISSING,
    default_factory: Callable[..., object] | object = dataclasses.MISSING,
    metadata: Mapping[Any, Any] | None = None,
    directives: Sequence[object] | None = (),
    extensions: list[FieldExtension] | None = None,
    graphql_type: Any | None = None,
    # This init parameter is used by PyRight to determine whether this field
    # is added in the constructor or not. It is not used to change
    # any behavior at the moment.
    init: Literal[True, False] | None = None,
) -> Any:
    """Annotates a method or property as a GraphQL mutation.

    Args:
        resolver: The resolver for the field. It can be a sync or async function.
        name: The GraphQL name of the field.
        description: The GraphQL description of the field.
        permission_classes: The permission classes required to access the field.
        deprecation_reason: The deprecation reason for the field.
        default: The default value for the field.
        default_factory: The default factory for the field.
        metadata: The metadata for the field.
        directives: The directives for the field.
        extensions: The extensions for the field.
        graphql_type: The GraphQL type for the field, useful when you want to use a
            different type in the resolver than the one in the schema.
        init: This parameter is used by PyRight to determine whether this field is
            added in the constructor or not. It is not used to change any behavior at
            the moment.

    Returns:
        The field object.

    This is normally used inside a type declaration:

    ```python
    import strawberry


    @strawberry.type
    class Mutation:
        @strawberry.mutation
        def create_post(self, title: str, content: str) -> Post: ...
    ```

    It can be used both as decorator and as a normal function.
    """
    return field(
        resolver=resolver,  # type: ignore
        name=name,
        description=description,
        permission_classes=permission_classes,
        deprecation_reason=deprecation_reason,
        default=default,
        default_factory=default_factory,
        metadata=metadata,
        directives=directives,
        extensions=extensions,
        graphql_type=graphql_type,
    )


# NOTE: we are separating the sync and async resolvers because using both
# in the same function will cause mypy to raise an error. Not sure if it is a bug


@overload
def subscription(
    *,
    resolver: _RESOLVER_TYPE_ASYNC[T],
    name: str | None = None,
    description: str | None = None,
    init: Literal[False] = False,
    permission_classes: list[type[BasePermission]] | None = None,
    deprecation_reason: str | None = None,
    default: Any = dataclasses.MISSING,
    default_factory: Callable[..., object] | object = dataclasses.MISSING,
    metadata: Mapping[Any, Any] | None = None,
    directives: Sequence[object] | None = (),
    extensions: list[FieldExtension] | None = None,
    graphql_type: Any | None = None,
) -> T: ...


@overload
def subscription(
    *,
    resolver: _RESOLVER_TYPE_SYNC[T],
    name: str | None = None,
    description: str | None = None,
    init: Literal[False] = False,
    permission_classes: list[type[BasePermission]] | None = None,
    deprecation_reason: str | None = None,
    default: Any = dataclasses.MISSING,
    default_factory: Callable[..., object] | object = dataclasses.MISSING,
    metadata: Mapping[Any, Any] | None = None,
    directives: Sequence[object] | None = (),
    extensions: list[FieldExtension] | None = None,
    graphql_type: Any | None = None,
) -> T: ...


@overload
def subscription(
    *,
    name: str | None = None,
    description: str | None = None,
    init: Literal[True] = True,
    permission_classes: list[type[BasePermission]] | None = None,
    deprecation_reason: str | None = None,
    default: Any = dataclasses.MISSING,
    default_factory: Callable[..., object] | object = dataclasses.MISSING,
    metadata: Mapping[Any, Any] | None = None,
    directives: Sequence[object] | None = (),
    extensions: list[FieldExtension] | None = None,
    graphql_type: Any | None = None,
) -> Any: ...


@overload
def subscription(
    resolver: _RESOLVER_TYPE_ASYNC[T],
    *,
    name: str | None = None,
    description: str | None = None,
    permission_classes: list[type[BasePermission]] | None = None,
    deprecation_reason: str | None = None,
    default: Any = dataclasses.MISSING,
    default_factory: Callable[..., object] | object = dataclasses.MISSING,
    metadata: Mapping[Any, Any] | None = None,
    directives: Sequence[object] | None = (),
    extensions: list[FieldExtension] | None = None,
    graphql_type: Any | None = None,
) -> StrawberryField: ...


@overload
def subscription(
    resolver: _RESOLVER_TYPE_SYNC[T],
    *,
    name: str | None = None,
    description: str | None = None,
    permission_classes: list[type[BasePermission]] | None = None,
    deprecation_reason: str | None = None,
    default: Any = dataclasses.MISSING,
    default_factory: Callable[..., object] | object = dataclasses.MISSING,
    metadata: Mapping[Any, Any] | None = None,
    directives: Sequence[object] | None = (),
    extensions: list[FieldExtension] | None = None,
    graphql_type: Any | None = None,
) -> StrawberryField: ...


def subscription(
    resolver: _RESOLVER_TYPE[Any] | None = None,
    *,
    name: str | None = None,
    description: str | None = None,
    permission_classes: list[type[BasePermission]] | None = None,
    deprecation_reason: str | None = None,
    default: Any = dataclasses.MISSING,
    default_factory: Callable[..., object] | object = dataclasses.MISSING,
    metadata: Mapping[Any, Any] | None = None,
    directives: Sequence[object] | None = (),
    extensions: list[FieldExtension] | None = None,
    graphql_type: Any | None = None,
    init: Literal[True, False] | None = None,
) -> Any:
    """Annotates a method or property as a GraphQL subscription.

    Args:
        resolver: The resolver for the field.
        name: The GraphQL name of the field.
        description: The GraphQL description of the field.
        permission_classes: The permission classes required to access the field.
        deprecation_reason: The deprecation reason for the field.
        default: The default value for the field.
        default_factory: The default factory for the field.
        metadata: The metadata for the field.
        directives: The directives for the field.
        extensions: The extensions for the field.
        graphql_type: The GraphQL type for the field, useful when you want to use a
            different type in the resolver than the one in the schema.
        init: This parameter is used by PyRight to determine whether this field is
            added in the constructor or not. It is not used to change any behavior at
            the moment.

    Returns:
        The field for the subscription.

    This is normally used inside a type declaration:

    ```python
    import strawberry


    @strawberry.type
    class Subscription:
        @strawberry.subscription
        def post_created(self, title: str, content: str) -> Post: ...
    ```

    it can be used both as decorator and as a normal function.
    """
    return field(
        resolver=resolver,  # type: ignore
        name=name,
        description=description,
        is_subscription=True,
        permission_classes=permission_classes,
        deprecation_reason=deprecation_reason,
        default=default,
        default_factory=default_factory,
        metadata=metadata,
        directives=directives,
        extensions=extensions,
        graphql_type=graphql_type,
    )


__all__ = ["mutation", "subscription"]



import re
from typing import Annotated, Any, ForwardRef, TypeVar

_parent_re = re.compile(r"^(?:strawberry\.)?Parent\[(.*)\]$")


class StrawberryParent: ...


T = TypeVar("T")

Parent = Annotated[T, StrawberryParent()]
"""Represents a parameter holding the parent resolver's value.

This can be used when defining a resolver on a type when the parent isn't expected
to return the type itself.

Example:

```python
import strawberry
from dataclasses import dataclass


@dataclass
class UserRow:
    id_: str


@strawberry.type
class User:
    @strawberry.field
    @staticmethod
    async def name(parent: strawberry.Parent[UserRow]) -> str:
        return f"User Number {parent.id_}"


@strawberry.type
class Query:
    @strawberry.field
    def user(self) -> User:
        return UserRow(id_="1234")
```
"""


def resolve_parent_forward_arg(annotation: Any) -> Any:
    if isinstance(annotation, str):
        str_annotation = annotation
    elif isinstance(annotation, ForwardRef):
        str_annotation = annotation.__forward_arg__
    else:
        # If neither, return the annotation as is
        return annotation

    if parent_match := _parent_re.match(str_annotation):
        annotation = Parent[ForwardRef(parent_match.group(1))]  # type: ignore[misc]

    return annotation


__all__ = ["Parent", "StrawberryParent", "resolve_parent_forward_arg"]



from __future__ import annotations

import sys
from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    Any,
    NewType,
    Optional,
    TypeVar,
    overload,
)

from strawberry.exceptions import InvalidUnionTypeError
from strawberry.types.base import StrawberryType
from strawberry.utils.str_converters import to_camel_case

if TYPE_CHECKING:
    from collections.abc import Callable, Iterable, Mapping

    from graphql import GraphQLScalarType


_T = TypeVar("_T", bound=type | NewType)


def identity(x: _T) -> _T:
    return x


@dataclass
class ScalarDefinition(StrawberryType):
    name: str
    description: str | None
    specified_by_url: str | None
    serialize: Callable | None
    parse_value: Callable | None
    parse_literal: Callable | None
    directives: Iterable[object] = ()
    origin: GraphQLScalarType | type | None = None

    # Optionally store the GraphQLScalarType instance so that we don't get
    # duplicates
    implementation: GraphQLScalarType | None = None

    # used for better error messages
    _source_file: str | None = None
    _source_line: int | None = None

    def copy_with(
        self, type_var_map: Mapping[str, StrawberryType | type]
    ) -> StrawberryType | type:
        return super().copy_with(type_var_map)  # type: ignore[safe-super]

    @property
    def is_graphql_generic(self) -> bool:
        return False


class ScalarWrapper:
    _scalar_definition: ScalarDefinition

    def __init__(self, wrap: Callable[[Any], Any]) -> None:
        self.wrap = wrap

    def __call__(self, *args: str, **kwargs: Any) -> Any:
        return self.wrap(*args, **kwargs)

    def __or__(self, other: StrawberryType | type) -> StrawberryType:
        if other is None:
            # Return the correct notation when using `StrawberryUnion | None`.
            return Optional[self]  # noqa: UP045

        # Raise an error in any other case.
        # There is Work in progress to deal with more merging cases, see:
        # https://github.com/strawberry-graphql/strawberry/pull/1455
        raise InvalidUnionTypeError(str(other), self.wrap)


def _process_scalar(
    cls: _T,
    *,
    name: str | None = None,
    description: str | None = None,
    specified_by_url: str | None = None,
    serialize: Callable | None = None,
    parse_value: Callable | None = None,
    parse_literal: Callable | None = None,
    directives: Iterable[object] = (),
) -> ScalarWrapper:
    from strawberry.exceptions.handler import should_use_rich_exceptions

    name = name or to_camel_case(cls.__name__)

    _source_file = None
    _source_line = None

    if should_use_rich_exceptions():
        frame = sys._getframe(3)

        _source_file = frame.f_code.co_filename
        _source_line = frame.f_lineno

    wrapper = ScalarWrapper(cls)
    wrapper._scalar_definition = ScalarDefinition(
        name=name,
        description=description,
        specified_by_url=specified_by_url,
        serialize=serialize,
        parse_literal=parse_literal,
        parse_value=parse_value,
        directives=directives,
        origin=cls,  # type: ignore[arg-type]
        _source_file=_source_file,
        _source_line=_source_line,
    )

    return wrapper


@overload
def scalar(
    cls: None = None,
    *,
    name: str,
    description: str | None = None,
    specified_by_url: str | None = None,
    serialize: Callable = identity,
    parse_value: Callable | None = None,
    parse_literal: Callable | None = None,
    directives: Iterable[object] = (),
) -> ScalarDefinition: ...


@overload
def scalar(
    cls: None = None,
    *,
    name: None = None,
    description: str | None = None,
    specified_by_url: str | None = None,
    serialize: Callable = identity,
    parse_value: Callable | None = None,
    parse_literal: Callable | None = None,
    directives: Iterable[object] = (),
) -> Callable[[_T], _T]: ...


@overload
def scalar(
    cls: _T,
    *,
    name: str | None = None,
    description: str | None = None,
    specified_by_url: str | None = None,
    serialize: Callable = identity,
    parse_value: Callable | None = None,
    parse_literal: Callable | None = None,
    directives: Iterable[object] = (),
) -> _T: ...


# TODO: We are tricking pyright into thinking that we are returning the given type
# here or else it won't let us use any custom scalar to annotate attributes in
# dataclasses/types. This should be properly solved when implementing StrawberryScalar
def scalar(
    cls: _T | None = None,
    *,
    name: str | None = None,
    description: str | None = None,
    specified_by_url: str | None = None,
    serialize: Callable = identity,
    parse_value: Callable | None = None,
    parse_literal: Callable | None = None,
    directives: Iterable[object] = (),
) -> Any:
    """Annotates a class or type as a GraphQL custom scalar.

    This function can be used in three ways:

    1. With a `name` but no `cls`: Returns a `ScalarDefinition` for use in
       `StrawberryConfig.scalar_map`. This is the recommended approach as it
       provides proper type checking support.

    2. As a decorator (no `cls`): Returns a decorator function. When the `cls`
       argument is provided inline, this is deprecated in favor of using
       `scalar_map`.

    3. With a `cls` argument (deprecated): Wraps the class/type directly.
       This approach is deprecated because it causes type checker issues.
       Use `scalar_map` in `StrawberryConfig` instead.

    Args:
        cls: The class or type to annotate (deprecated, use scalar_map instead).
        name: The GraphQL name of the scalar.
        description: The description of the scalar.
        specified_by_url: The URL of the specification.
        serialize: The function to serialize the scalar.
        parse_value: The function to parse the value.
        parse_literal: The function to parse the literal.
        directives: The directives to apply to the scalar.

    Returns:
        A `ScalarDefinition` when called with `name` only, a decorator function
        when called without arguments, or the wrapped type when called with `cls`.

    Example usages:

    Recommended approach using scalar_map:

    ```python
    from typing import NewType
    import strawberry
    from strawberry.schema.config import StrawberryConfig

    # Define the type
    Base64 = NewType("Base64", bytes)

    # Configure the scalar in schema config
    schema = strawberry.Schema(
        query=Query,
        config=StrawberryConfig(
            scalar_map={
                Base64: strawberry.scalar(
                    name="Base64",
                    serialize=lambda v: base64.b64encode(v).decode(),
                    parse_value=lambda v: base64.b64decode(v),
                )
            }
        ),
    )
    ```

    Legacy approach (deprecated):

    ```python
    Base64Encoded = strawberry.scalar(
        NewType("Base64Encoded", bytes),
        serialize=base64.b64encode,
        parse_value=base64.b64decode,
    )
    ```
    """
    from strawberry.exceptions.handler import should_use_rich_exceptions

    _source_file = None
    _source_line = None

    if should_use_rich_exceptions():
        frame = sys._getframe(1)
        _source_file = frame.f_code.co_filename
        _source_line = frame.f_lineno

    if cls is None and name is not None:
        return ScalarDefinition(
            name=name,
            description=description,
            specified_by_url=specified_by_url,
            serialize=serialize,
            parse_literal=parse_literal,
            parse_value=parse_value,
            directives=directives,
            origin=None,
            _source_file=_source_file,
            _source_line=_source_line,
        )

    if parse_value is None:
        parse_value = cls

    def wrap(cls: _T) -> ScalarWrapper:
        import warnings

        warnings.warn(
            "Passing a class to strawberry.scalar() is deprecated. "
            "Use StrawberryConfig.scalar_map instead for better type checking support. "
            "See: https://strawberry.rocks/docs/types/scalars",
            DeprecationWarning,
            stacklevel=3,
        )
        return _process_scalar(
            cls,
            name=name,
            description=description,
            specified_by_url=specified_by_url,
            serialize=serialize,
            parse_value=parse_value,
            parse_literal=parse_literal,
            directives=directives,
        )

    if cls is None:
        return wrap

    return wrap(cls)


__all__ = ["ScalarDefinition", "ScalarWrapper", "scalar"]



from collections.abc import Callable, Iterator
from functools import cache, lru_cache
from typing import Any

from graphql.language.parser import parse

from strawberry.extensions.base_extension import SchemaExtension


@cache
def _get_parse_cache(maxsize: int | None) -> Callable[..., Any]:
    # Shared LRU caches keyed by ``maxsize``. Multiple ``ParserCache`` instances
    # (e.g. a fresh one constructed per request via the factory pattern) reuse
    # the same wrapped ``parse`` so caching is effective across requests
    # without sharing extension instances.
    return lru_cache(maxsize=maxsize)(parse)


class ParserCache(SchemaExtension):
    """Add LRU caching the parsing step during execution to improve performance.

    Pass it as a factory; the LRU cache lives at module level and is keyed by
    ``maxsize``, so it is shared across every request and every schema that
    constructs a ``ParserCache`` with the same ``maxsize``.

    ```python
    import strawberry
    from strawberry.extensions import ParserCache

    schema = strawberry.Schema(
        Query,
        extensions=[lambda: ParserCache(maxsize=100)],
    )
    ```
    """

    def __init__(self, maxsize: int | None = None) -> None:
        """Initialize the ParserCache.

        Args:
            maxsize: Set the maxsize of the cache. If `maxsize` is set to `None` then the
                cache will grow without bound.
                More info: https://docs.python.org/3/library/functools.html#functools.lru_cache
        """
        super().__init__()
        self.cached_parse_document = _get_parse_cache(maxsize)

    def on_parse(self) -> Iterator[None]:
        execution_context = self.execution_context

        execution_context.graphql_document = self.cached_parse_document(
            execution_context.query, **execution_context.parse_options
        )
        yield


__all__ = ["ParserCache"]
